@@ -18,6 +18,8 @@
 
 #define _KMEMUSER
 
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/stat.h>
@@ -101,7 +103,7 @@ char *cpustatenames[] = {
 
 /* these are for detailing the memory statistics */
 
-int memory_stats[5];
+long memory_stats[5];
 char *memorynames[] = {
     "K max, ", "K avail, ", "K free, ", "K swap, ", "K free swap", NULL
 };
@@ -114,6 +116,9 @@ extern char *format_k();
 extern char *format_time();
 extern long percentages();
 
+/* forward references */
+int proc_compare (void *pp1, void *pp2);
+
 #define X_AVENRUN	0
 #define X_NPROC		1
 #define X_FREEMEM	2
@@ -325,10 +330,10 @@ void get_system_info(si)
   return;
 }
 
-caddr_t get_process_info(si, sel, compare)
+caddr_t get_process_info(si, sel, x)
      struct system_info *si;
      struct process_select *sel;
-     int (*compare)();
+     int x;
 {
   register int i;
   register int total_procs;
@@ -380,7 +385,7 @@ caddr_t get_process_info(si, sel, compare)
 
   /* if requested, sort the "interesting" processes */
   if (compare != NULL)
-    qsort ((char *) pref, active_procs, sizeof (struct prpsinfo *), compare);
+    qsort ((char *) pref, active_procs, sizeof (struct prpsinfo *), proc_compare);
 
   /* remember active and total counts */
   si->p_total = total_procs;
@@ -461,7 +466,7 @@ int getkval(offset, ptr, size, refstr)
       refstr++;
     (void) fprintf(stderr, "%s: lseek to %s: %s\n", KMEM,
 		   refstr, strerror(errno));
-    exit(0);
+    quit(0);
   }
   if (read(kmem, (char *) ptr, size) == -1) {
     if (*refstr == '!')
@@ -469,7 +474,7 @@ int getkval(offset, ptr, size, refstr)
     else {
       (void) fprintf(stderr, "%s: reading %s: %s\n", KMEM,
 		     refstr, strerror(errno));
-      exit(0);
+      quit(0);
     }
   }
   return(1);
@@ -500,16 +505,16 @@ unsigned char sorted_state[] =
 };
 
 int proc_compare (pp1, pp2)
-     struct prpsinfo **pp1;
-     struct prpsinfo **pp2;
+     void *pp1;
+     void *pp2;
 {
   register struct prpsinfo *p1;
   register struct prpsinfo *p2;
   register long result;
 
   /* remove one level of indirection */
-  p1 = *pp1;
-  p2 = *pp2;
+  p1 = *(struct prpsinfo **)pp1;
+  p2 = *(struct prpsinfo **)pp2;
 
   /* compare percent cpu (pctcpu) */
   if ((result = (long) (p2->pr_cpu - p1->pr_cpu)) == 0) {
@@ -533,15 +538,15 @@ int proc_compare (pp1, pp2)
 }
 
 /* return the owner of the specified process. */
-uid_t proc_owner (pid)
-     pid_t pid;
+int proc_owner (pid)
+     int pid;
 {
   register struct prpsinfo *p;
   int i;
 
   for (i = 0, p = pbase; i < nproc; i++, p++)
-    if (p->pr_pid == pid)
-      return (p->pr_uid);
+    if (p->pr_pid == (oid_t)pid)
+      return ((int)p->pr_uid);
 
   return (-1);
 }
