@@ -14,6 +14,7 @@
  *	SunOS 4.1.1
  *	SunOS 4.1.2 (including MP architectures)
  *	SunOS 4.1.3 (including MP architectures)
+ *      SunOS 4.1.3_U1 (including MP architectures)
  *      SunOS 4.1.4 (including MP architectures)
  *	Solbourne OS/MP PRIOR to 4.1A
  *
@@ -21,7 +22,7 @@
  *
  * CFLAGS: -DHAVE_GETOPT -DORDER
  *
- * AUTHOR:  William LeFebvre <phil@eecs.nwu.edu>
+ * AUTHOR:  William LeFebvre <wnl@groupsys.com>
  * Solbourne support by David MacKenzie <djm@eng.umd.edu>
  */
 
@@ -30,6 +31,7 @@
  * #ifdef solbourne is for Solbourne.
  */
 
+#include "config.h"
 #include <sys/types.h>
 #include <sys/signal.h>
 
@@ -207,26 +209,24 @@ char *cpustatenames[] = {
 
 /* these are for detailing the memory statistics */
 
-int memory_stats[4];
+long memory_stats[4];
 char *memorynames[] = {
     "K available, ", "K in use, ", "K free, ", "K locked", NULL
 };
 
 /* these are names given to allowed sorting orders -- first is default */
 char *ordernames[] = 
-{"cpu", "size", "res", "time", NULL};
+{"cpu", "size", "res", NULL};
 
 /* forward definitions for comparison functions */
 int compare_cpu();
 int compare_size();
 int compare_res();
-int compare_time();
 
 int (*proc_compares[])() = {
     compare_cpu,
     compare_size,
     compare_res,
-    compare_time,
     NULL };
 
 
@@ -381,9 +381,7 @@ struct statics *statics;
     statics->procstate_names = procstatenames;
     statics->cpustate_names = cpustatenames;
     statics->memory_names = memorynames;
-#ifdef ORDER
     statics->order_names = ordernames;
-#endif
 
     /* all done! */
     return(0);
@@ -405,6 +403,7 @@ register char *uname_field;
     return(header);
 }
 
+void
 get_system_info(si)
 
 struct system_info *si;
@@ -532,11 +531,11 @@ struct system_info *si;
 
 static struct handle handle;
 
-caddr_t get_process_info(si, sel, compare)
+caddr_t get_process_info(si, sel, compare_index)
 
 struct system_info *si;
 struct process_select *sel;
-int (*compare)();
+int compare_index;
 
 {
     register int i;
@@ -592,10 +591,8 @@ int (*compare)();
     }
 
     /* if requested, sort the "interesting" processes */
-    if (compare != NULL)
-    {
-	qsort((char *)pref, active_procs, sizeof(struct proc *), compare);
-    }
+    qsort((char *)pref, active_procs, sizeof(struct proc *),
+	  proc_compares[compare_index]);
 
     /* remember active and total counts */
     si->p_total = total_procs;
@@ -793,8 +790,6 @@ char *refstr;
     return(1);
 }
     
-/* comparison routine for qsort */
-
 /* comparison routines for qsort */
 
 /*
@@ -920,35 +915,6 @@ struct proc **pp2;
     return(result);
 }
 
-/* compare_time - the comparison function for sorting by total cpu time */
-
-compare_time(pp1, pp2)
-
-struct proc **pp1;
-struct proc **pp2;
-
-{
-    register struct proc *p1;
-    register struct proc *p2;
-    register int result;
-    register pctcpu lresult;
-
-    /* remove one level of indirection */
-    p1 = *pp1;
-    p2 = *pp2;
-
-    ORDERKEY_CPTICKS
-    ORDERKEY_PCTCPU
-    ORDERKEY_STATE
-    ORDERKEY_PRIO
-    ORDERKEY_RSSIZE
-    ORDERKEY_MEM
-    ;
-
-    return(result);
-}
-
-
 /*
  * proc_owner(pid) - returns the uid that owns process "pid", or -1 if
  *		the process does not exist.
