@@ -1,5 +1,5 @@
 char *copyright =
-    "Copyright (c) 1984 through 1996, William LeFebvre";
+    "Copyright (c) 1984 through 2006, William LeFebvre";
 
 /*
  *  Top users/processes display for Unix
@@ -9,7 +9,9 @@ char *copyright =
  *  but this entire comment MUST remain intact.
  *
  *  Copyright (c) 1984, 1989, William LeFebvre, Rice University
- *  Copyright (c) 1989, 1990, 1992, William LeFebvre, Northwestern University
+ *  Copyright (c) 1989 - 1994, William LeFebvre, Northwestern University
+ *  Copyright (c) 1994, 1995, William LeFebvre, Argonne National Laboratory
+ *  Copyright (c) 1996, William LeFebvre, Group sys Consulting
  */
 
 /*
@@ -24,8 +26,6 @@ char *copyright =
  * The following preprocessor variables, when defined, are used to
  * distinguish between different Unix implementations:
  *
- *	SIGHOLD  - use SVR4 sighold function when defined
- *	SIGRELSE - use SVR4 sigrelse function when defined
  *	FD_SET   - macros FD_SET and FD_ZERO are used when defined
  */
 
@@ -33,16 +33,30 @@ char *copyright =
 #include <signal.h>
 #include <setjmp.h>
 #include <ctype.h>
-#include <sys/time.h>
+
+/* determine which type of signal functions to use */
+#ifdef HAVE_SIGACTION
+#undef HAVE_SIGHOLD
+#else
+#if !defined(HAVE_SIGHOLD) || !defined(HAVE_SIGRELSE)
+#define BSD_SIGNALS
+#endif
+#endif
 
 /* includes specific to top */
+
+#include "top.h"
+#include "machine.h"
+#include "commands.h"
 #include "display.h"		/* interface to display package */
 #include "screen.h"		/* interface to screen package */
-#include "top.h"
-#include "top.local.h"
 #include "boolean.h"
-#include "machine.h"
+#include "username.h"
 #include "utils.h"
+#include "version.h"
+#ifdef ENABLE_COLOR
+#include "color.h"
+#endif
 
 /* Size of the stdio buffer given to stdout */
 #define Buffersize	2048
@@ -50,11 +64,10 @@ char *copyright =
 /* The buffer that stdio will use */
 char stdoutbuf[Buffersize];
 
-/* build Signal masks */
-#define Smask(s)	(1 << ((s) - 1))
-
-/* for system errors */
-extern int errno;
+/* build signal masks */
+#ifndef sigmask
+#define sigmask(s)	(1 << ((s) - 1))
+#endif
 
 /* for getopt: */
 extern int  optind;
@@ -63,17 +76,6 @@ extern char *optarg;
 /* imported from screen.c */
 extern int overstrike;
 
-/* signal handling routines */
-sigret_t leave();
-sigret_t onalrm();
-sigret_t tstop();
-#ifdef SIGWINCH
-sigret_t winch();
-#endif
-
-/* internal routines */
-void quit();
-
 /* values which need to be accessed by signal handlers */
 static int max_topn;		/* maximum displayable processes */
 
@@ -81,57 +83,152 @@ static int max_topn;		/* maximum displayable processes */
 char *myname = "top";
 jmp_buf jmp_int;
 
-/* routines that don't return int */
-
-char *username();
-char *ctime();
-char *kill_procs();
-char *renice_procs();
-
-#ifdef ORDER
-extern int (*proc_compares[])();
-#else
-extern int proc_compare();
-#endif
-time_t time();
-
-caddr_t get_process_info();
-
-/* different routines for displaying the user's identification */
-/* (values assigned to get_userid) */
-char *username();
-char *itoa7();
-
-/* display routines that need to be predeclared */
-int i_loadave();
-int u_loadave();
-int i_procstates();
-int u_procstates();
-int i_cpustates();
-int u_cpustates();
-int i_memory();
-int u_memory();
-int i_message();
-int u_message();
-int i_header();
-int u_header();
-int i_process();
-int u_process();
-
 /* pointers to display routines */
-int (*d_loadave)() = i_loadave;
-int (*d_procstates)() = i_procstates;
-int (*d_cpustates)() = i_cpustates;
-int (*d_memory)() = i_memory;
-int (*d_message)() = i_message;
-int (*d_header)() = i_header;
-int (*d_process)() = i_process;
+void (*d_loadave)(int, double *) = i_loadave;
+void (*d_minibar)(int (*)(char *, int)) = i_minibar;
+void (*d_uptime)(time_t *, time_t *) = i_uptime;
+void (*d_procstates)(int, int*) = i_procstates;
+void (*d_cpustates)(int *) = i_cpustates;
+void (*d_memory)(long *) = i_memory;
+void (*d_swap)(long *) = i_swap;
+void (*d_message)() = i_message;
+void (*d_header)(char *) = i_header;
+void (*d_process)(int, char *) = i_process;
 
+/*
+ *  reset_display() - reset all the display routine pointers so that entire
+ *	screen will get redrawn.
+ */
 
-main(argc, argv)
+void
+reset_display()
 
-int  argc;
-char *argv[];
+{
+    d_loadave    = i_loadave;
+    d_minibar    = i_minibar;
+    d_uptime     = i_uptime;
+    d_procstates = i_procstates;
+    d_cpustates  = i_cpustates;
+    d_memory     = i_memory;
+    d_swap       = i_swap;
+    d_message	 = i_message;
+    d_header	 = i_header;
+    d_process	 = i_process;
+}
+
+/*
+ *  signal handlers
+ */
+
+void
+set_signal(int sig, RETSIGTYPE (*handler)(int))
+
+{
+#ifdef HAVE_SIGACTION
+    struct sigaction action;
+
+    action.sa_handler = handler;
+    action.sa_flags = 0;
+    (void) sigaction(sig, &action, NULL);
+#else
+    (void) signal(sig, handler);
+#endif
+}
+
+RETSIGTYPE
+leave(int i)	/* exit under normal conditions -- INT handler */
+
+{
+    end_screen();
+    exit(0);
+}
+
+RETSIGTYPE
+tstop(int i)	/* SIGTSTP handler */
+
+{
+#ifdef HAVE_SIGACTION
+    sigset_t set;
+    struct sigaction sa;
+#endif
+
+    /* move to the lower left */
+    end_screen();
+    fflush(stdout);
+
+    /* default the signal handler action */
+    set_signal(SIGTSTP, SIG_DFL);
+
+    /* unblock the TSTP signal */
+#ifdef HAVE_SIGACTION
+    sigemptyset(&set);
+    sigaddset(&set, SIGTSTP);
+    sigprocmask(SIG_UNBLOCK, &set, NULL);
+#endif
+
+#ifdef HAVE_SIGHOLD
+    sigrelse(SIGTSTP);
+#endif
+
+#ifdef BSD_SIGNALS
+    (void) sigsetmask(sigblock(0) & ~(sigmask(SIGTSTP)));
+#endif
+
+    /* send ourselves a TSTP to stop the process */
+    (void) kill(0, SIGTSTP);
+
+    /* reset the signal handler */
+    set_signal(SIGTSTP, tstop);
+
+    /* reinit screen */
+    reinit_screen();
+
+    /* jump to appropriate place */
+    longjmp(jmp_int, 1);
+
+    /* NOTREACHED */
+}
+
+#ifdef SIGWINCH
+RETSIGTYPE
+winch(int i)		/* SIGWINCH handler */
+
+{
+    /* reascertain the screen dimensions */
+    get_screensize();
+
+    /* tell display to resize */
+    max_topn = display_resize();
+
+#ifndef HAVE_SIGACTION
+    /* reset the signal handler */
+    set_signal(SIGWINCH, winch);
+#endif
+
+    /* jump to appropriate place */
+    longjmp(jmp_int, 1);
+}
+#endif
+
+void
+quit(int status)		/* exit under duress */
+
+{
+    end_screen();
+    exit(status);
+    /*NOTREACHED*/
+}
+
+RETSIGTYPE
+onalrm(int i)	/* SIGALRM handler */
+
+{
+    /* this is only used in batch mode to break out of the pause() */
+    /* return; */
+}
+
+int
+main(int argc, char *argv[])
 
 {
     register int i;
@@ -149,7 +246,7 @@ char *argv[];
     int delay = Default_DELAY;
     int displays = 0;		/* indicates unspecified */
     time_t curr_time;
-    char *(*get_userid)() = username;
+    char *(*get_userid)(int) = username;
     char *uname_field = "USERNAME";
     char *header_text;
     char *env_top;
@@ -160,7 +257,7 @@ char *argv[];
     char dostates = No;
     char do_unames = Yes;
     char interactive = Maybe;
-    char warnings = 0;
+    char show_tags = No;
 #if Default_TOPN == Infinity
     char topn_specified = No;
 #endif
@@ -169,23 +266,25 @@ char *argv[];
     char no_command = 1;
     struct timeval timeout;
     struct process_select ps;
-#ifdef ORDER
     char *order_name = NULL;
     int order_index = 0;
-#endif
 #ifndef FD_SET
     /* FD_SET and friends are not present:  fake it */
     typedef int fd_set;
 #define FD_ZERO(x)     (*(x) = 0)
-#define FD_SET(f, x)   (*(x) = f)
+#define FD_SET(f, x)   (*(x) = 1<<f)
 #endif
     fd_set readfds;
-
-#ifdef ORDER
-    static char command_chars[] = "\f qh?en#sdkriIuo";
-#else
-    static char command_chars[] = "\f qh?en#sdkriIu";
+    static struct ext_decl exts = { NULL, NULL };
+#ifdef ENABLE_COLOR
+    int color_on = 1;
 #endif
+#ifdef HAVE_SIGPROCMASK
+    sigset_t signalset;
+#endif
+
+    static char command_chars[] = "\f qh?en#sdkriIucoCNPMT";
+
 /* these defines enumerate the "strchr"s of the commands in command_chars */
 #define CMD_redraw	0
 #define CMD_update	1
@@ -203,16 +302,16 @@ char *argv[];
 #define CMD_idletog     12
 #define CMD_idletog2    13
 #define CMD_user	14
-#ifdef ORDER
-#define CMD_order       15
-#endif
+#define CMD_cmdline     15
+#define CMD_order       16
+#define CMD_color       17
+#define CMD_order_pid   18
+#define CMD_order_cpu   19
+#define CMD_order_mem   20
+#define CMD_order_time  21
 
     /* set the buffer for stdout */
-#ifdef DEBUG
-    setbuffer(stdout, NULL, 0);
-#else
     setbuffer(stdout, stdoutbuf, Buffersize);
-#endif
 
     /* get our name */
     if (argc > 0)
@@ -230,6 +329,7 @@ char *argv[];
     /* initialize some selection options */
     ps.idle    = Yes;
     ps.system  = No;
+    ps.fullcmd = No;
     ps.uid     = -1;
     ps.command = NULL;
 
@@ -256,15 +356,31 @@ char *argv[];
 	    optind = 1;
 	}
 
-	while ((i = getopt(ac, av, "SIbinqus:d:U:o:")) != EOF)
+	while ((i = getopt(ac, av, "CDSITbcinquvs:d:U:o:")) != EOF)
 	{
 	    switch(i)
 	    {
-	      case 'u':			/* toggle uid/username display */
+#ifdef ENABLE_COLOR
+	    case 'C':
+		color_on = !color_on;
+		break;
+#endif
+
+	    case 'D':
+		debug_set(1);
+		break;
+
+	    case 'v':			/* show version number */
+		fprintf(stderr, "%s: version %s\n",
+			myname, version_string());
+		exit(1);
+		break;
+
+	    case 'u':			/* toggle uid/username display */
 		do_unames = !do_unames;
 		break;
 
-	      case 'U':			/* display only username's processes */
+	    case 'U':			/* display only username's processes */
 		if ((ps.uid = userid(optarg)) == -1)
 		{
 		    fprintf(stderr, "%s: unknown user\n", optarg);
@@ -272,30 +388,36 @@ char *argv[];
 		}
 		break;
 
-	      case 'S':			/* show system processes */
+	    case 'S':			/* show system processes */
 		ps.system = !ps.system;
 		break;
 
-	      case 'I':                   /* show idle processes */
+	    case 'I':                   /* show idle processes */
 		ps.idle = !ps.idle;
 		break;
 
-	      case 'i':			/* go interactive regardless */
+	    case 'T':			/* show color tags */
+		show_tags = 1;
+		break;
+
+	    case 'i':			/* go interactive regardless */
 		interactive = Yes;
 		break;
 
-	      case 'n':			/* batch, or non-interactive */
-	      case 'b':
+	    case 'c':
+		ps.fullcmd = Yes;
+		break;
+
+	    case 'n':			/* batch, or non-interactive */
+	    case 'b':
 		interactive = No;
 		break;
 
-	      case 'd':			/* number of displays to show */
+	    case 'd':			/* number of displays to show */
 		if ((i = atoiwi(optarg)) == Invalid || i == 0)
 		{
-		    fprintf(stderr,
-			"%s: warning: display count should be positive -- option ignored\n",
-			myname);
-		    warnings++;
+		    new_message(MT_standout | MT_delayed,
+				" Bad display count (ignored)");
 		}
 		else
 		{
@@ -303,18 +425,16 @@ char *argv[];
 		}
 		break;
 
-	      case 's':
-		if ((delay = atoi(optarg)) < 0)
+	    case 's':
+		if ((delay = atoi(optarg)) < 0 || (delay == 0 && getuid() != 0))
 		{
-		    fprintf(stderr,
-			"%s: warning: seconds delay should be non-negative -- using default\n",
-			myname);
+		    new_message(MT_standout | MT_delayed,
+				" Bad seconds delay (ignored)");
 		    delay = Default_DELAY;
-		    warnings++;
 		}
 		break;
 
-	      case 'q':		/* be quick about it */
+	    case 'q':		/* be quick about it */
 		/* only allow this if user is really root */
 		if (getuid() == 0)
 		{
@@ -323,56 +443,46 @@ char *argv[];
 		}
 		else
 		{
-		    fprintf(stderr,
-			"%s: warning: `-q' option can only be used by root\n",
-			myname);
-		    warnings++;
+		    new_message(MT_standout | MT_delayed,
+				" Option -q can only be used by root");
 		}
 		break;
 
-	      case 'o':		/* select sort order */
-#ifdef ORDER
+	    case 'o':		/* select sort order */
 		order_name = optarg;
-#else
-		fprintf(stderr,
-			"%s: this platform does not support arbitrary ordering.  Sorry.\n",
-			myname);
-		warnings++;
-#endif
 		break;
 
-	      default:
+	    default:
 		fprintf(stderr, "\
 Top version %s\n\
-Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
+Usage: %s [-ISTbcinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 			version_string(), myname);
 		exit(1);
 	    }
 	}
 
 	/* get count of top processes to display (if any) */
-	if (optind < ac)
+	if (optind < ac && *av[optind])
 	{
-	    if ((topn = atoiwi(av[optind])) == Invalid)
+	    if ((i = atoiwi(av[optind])) == Invalid)
 	    {
-		fprintf(stderr,
-			"%s: warning: process display count should be non-negative -- using default\n",
-			myname);
-		warnings++;
+		new_message(MT_standout | MT_delayed,
+			    " Process count not a number (ignored)");
 	    }
-#if Default_TOPN == Infinity
             else
 	    {
+#if Default_TOPN == Infinity
 		topn_specified = Yes;
-	    }
 #endif
+		topn = i;
+	    }
 	}
 
 	/* tricky:  remember old value of preset_argc & set preset_argc = 0 */
 	i = preset_argc;
 	preset_argc = 0;
 
-    /* repeat only if we really did the preset arguments */
+	/* repeat only if we really did the preset arguments */
     } while (i != 0);
 
     /* set constants for username/uid display correctly */
@@ -382,17 +492,39 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	get_userid = itoa7;
     }
 
-    /* initialize the kernel memory interface */
+    /* in order to support forward compatability, we have to ensure that
+       the entire statics structure is set to a known value before we call
+       machine_init.  This way fields that a module does not know about
+       will retain their default values */
+    memzero((void *)&statics, sizeof(statics));
+    statics.boottime = -1;
+
+#ifdef ENABLE_COLOR
+    /* If colour has been turned on read in the settings. */
+    env_top = getenv("TOPCOLOURS");
+    if (!env_top)
+    {
+	env_top = getenv("TOPCOLORS");
+    }
+    /* must do something about error messages */
+    color_env_parse(env_top);
+#endif
+
+    /* call the platform-specific init */
     if (machine_init(&statics) == -1)
     {
 	exit(1);
     }
 
-#ifdef ORDER
     /* determine sorting order index, if necessary */
     if (order_name != NULL)
     {
-	if ((order_index = string_index(order_name, statics.order_names)) == -1)
+	if (statics.order_names == NULL)
+	{
+	    new_message(MT_standout | MT_delayed,
+			" This platform does not support arbitrary ordering");
+	}
+	else if ((order_index = string_index(order_name, statics.order_names)) == -1)
 	{
 	    char **pp;
 
@@ -408,14 +540,10 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	    exit(1);
 	}
     }
-#endif
 
-#ifdef no_initialization_needed
-    /* initialize the hashing stuff */
-    if (do_unames)
-    {
-	init_hash();
-    }
+#ifdef WITH_EXT
+    /* initialize extensions */
+    init_ext(&exts);
 #endif
 
     /* initialize termcap */
@@ -424,20 +552,34 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
     /* get the string to use for the process area header */
     header_text = format_header(uname_field);
 
+#ifdef ENABLE_COLOR
+    /* Disable colours on non-smart terminals */
+    if (!smart_terminal)
+    {
+	color_on = 0;
+    }
+#endif
+
     /* initialize display interface */
     if ((max_topn = display_init(&statics)) == -1)
     {
 	fprintf(stderr, "%s: can't allocate sufficient memory\n", myname);
 	exit(4);
     }
+
+    /* handle request for color tags */
+    if (show_tags)
+    {
+	color_dump(stdout);
+	exit(0);
+    }
     
     /* print warning if user requested more processes than we can display */
     if (topn > max_topn)
     {
-	fprintf(stderr,
-		"%s: warning: this terminal can only display %d processes.\n",
-		myname, max_topn);
-	warnings++;
+	new_message(MT_standout | MT_delayed,
+		    " This terminal can only display %d processes.",
+		    max_topn);
     }
 
     /* adjust for topn == Infinity */
@@ -455,7 +597,7 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	 */
 #if Default_TOPN == Infinity
 	topn = smart_terminal ? Largest :
-		    (topn_specified ? Largest : Nominal_TOPN);
+	    (topn_specified ? Largest : Nominal_TOPN);
 #else
 	topn = Largest;
 #endif
@@ -477,34 +619,42 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
     }
 
     /* hold interrupt signals while setting up the screen and the handlers */
-#ifdef SIGHOLD
+#ifdef HAVE_SIGPROCMASK
+    sigemptyset(&signalset);
+    sigaddset(&signalset, SIGINT);
+    sigaddset(&signalset, SIGQUIT);
+    sigaddset(&signalset, SIGTSTP);
+#ifdef SIGWINCH
+    sigaddset(&signalset, SIGWINCH);
+#endif
+    sigprocmask(SIG_BLOCK, &signalset, NULL);
+#endif
+
+#ifdef HAVE_SIGHOLD
     sighold(SIGINT);
     sighold(SIGQUIT);
     sighold(SIGTSTP);
-#else
-    old_sigmask = sigblock(Smask(SIGINT) | Smask(SIGQUIT) | Smask(SIGTSTP));
-#endif
-    init_screen();
-    (void) signal(SIGINT, leave);
-    (void) signal(SIGQUIT, leave);
-    (void) signal(SIGTSTP, tstop);
 #ifdef SIGWINCH
-    (void) signal(SIGWINCH, winch);
+    sighold(SIGWINCH);
 #endif
-#ifdef SIGRELSE
-    sigrelse(SIGINT);
-    sigrelse(SIGQUIT);
-    sigrelse(SIGTSTP);
+#endif
+
+#ifdef BSD_SIGNALS
+#ifdef SIGWINCH
+    old_sigmask = sigblock(sigmask(SIGINT) | sigmask(SIGQUIT) |
+			   sigmask(SIGTSTP) | sigmask(SIGWINCH));
 #else
-    (void) sigsetmask(old_sigmask);
+    old_sigmask = sigblock(sigmask(SIGINT) | sigmask(SIGQUIT) | sigmask(SIGTSTP));
+#endif
+#endif
+
+    init_screen();
+    (void) set_signal(SIGINT, leave);
+    (void) set_signal(SIGQUIT, leave);
+    (void) set_signal(SIGTSTP, tstop);
+#ifdef SIGWINCH
+    (void) set_signal(SIGWINCH, winch);
 #endif
-    if (warnings)
-    {
-	fputs("....", stderr);
-	fflush(stderr);			/* why must I do this? */
-	sleep((unsigned)(3 * warnings));
-	fputc('\n', stderr);
-    }
 
     /* setup the jump buffer for stops */
     if (setjmp(jmp_int) != 0)
@@ -513,6 +663,40 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	reset_display();
     }
 
+    /*
+     * Ready to release the signals.  This will also happen (needlessly)
+     * after a longjmp, but that's okay.
+     */
+#ifdef HAVE_SIGPROCMASK
+    sigprocmask(SIG_UNBLOCK, &signalset, NULL);
+#endif
+
+#ifdef HAVE_SIGHOLD
+    sigrelse(SIGINT);
+    sigrelse(SIGQUIT);
+    sigrelse(SIGTSTP);
+#ifdef SIGWINCH
+    sigrelse(SIGWINCH);
+#endif
+#endif
+
+#ifdef BSD_SIGNALS
+    (void) sigsetmask(old_sigmask);
+#endif
+
+    /* some systems require a warmup */
+    if (statics.flags.warmup)
+    {
+	get_system_info(&system_info);
+	(void)get_process_info(&system_info, &ps, 0);
+	timeout.tv_sec = 1;
+	timeout.tv_usec = 0;
+	select(0, NULL, NULL, NULL, &timeout);
+
+	/* if we've warmed up, then we can show good states too */
+	dostates = Yes;
+    }
+
     /*
      *  main loop -- repeat while display count is positive or while it
      *		indicates infinity (by being -1)
@@ -525,21 +709,28 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 
 	/* get the current set of processes */
 	processes =
-		get_process_info(&system_info,
-				 &ps,
-#ifdef ORDER
-				 proc_compares[order_index]);
-#else
-				 proc_compare);
-#endif
+	    get_process_info(&system_info,
+			     &ps,
+			     order_index);
 
 	/* display the load averages */
 	(*d_loadave)(system_info.last_pid,
 		     system_info.load_avg);
 
-	/* display the current time */
 	/* this method of getting the time SHOULD be fairly portable */
 	time(&curr_time);
+
+	/* if we have a minibar extension, use it, otherwise show uptime */
+	if (exts.f_minibar != NULL)
+	{
+	    (*d_minibar)(exts.f_minibar);
+	}
+	else
+	{
+	    (*d_uptime)(&statics.boottime, &curr_time);
+	}
+
+	/* display the current time */
 	i_timeofday(&curr_time);
 
 	/* display process state breakdown */
@@ -558,16 +749,15 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	    {
 		z_cpustates();
 	    }
-	    else
-	    {
-		putchar('\n');
-	    }
 	    dostates = Yes;
 	}
 
 	/* display memory stats */
 	(*d_memory)(system_info.memory);
 
+	/* display swap stats */
+	(*d_swap)(system_info.swap);
+
 	/* handle message area */
 	(*d_message)();
 
@@ -579,7 +769,7 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	    /* determine number of processes to actually display */
 	    /* this number will be the smallest of:  active processes,
 	       number user requested, number current screen accomodates */
-	    active_procs = system_info.p_active;
+	    active_procs = system_info.P_ACTIVE;
 	    if (active_procs > topn)
 	    {
 		active_procs = topn;
@@ -604,7 +794,13 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 	u_endscreen(i);
 
 	/* now, flush the output buffer */
-	fflush(stdout);
+	if (fflush(stdout) != 0)
+	{
+	    new_message(MT_standout, " Write error on stdout");
+	    putchar('\r');
+	    quit(1);
+	    /*NOTREACHED*/
+	}
 
 	/* only do the rest if we have more displays to show */
 	if (displays)
@@ -619,9 +815,12 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 		else
 		{
 		    d_loadave = u_loadave;
+		    d_minibar = u_minibar;
+		    d_uptime = u_uptime;
 		    d_procstates = u_procstates;
 		    d_cpustates = u_cpustates;
 		    d_memory = u_memory;
+		    d_swap = u_swap;
 		    d_message = u_message;
 		    d_header = u_header;
 		    d_process = u_process;
@@ -645,7 +844,7 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 
 		/* set up arguments for select with timeout */
 		FD_ZERO(&readfds);
-		FD_SET(1, &readfds);		/* for standard input */
+		FD_SET(0, &readfds);		/* for standard input */
 		timeout.tv_sec  = delay;
 		timeout.tv_usec = 0;
 
@@ -660,7 +859,14 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 
 		    /* now read it and convert to command strchr */
 		    /* (use "change" as a temporary to hold strchr) */
-		    (void) read(0, &ch, 1);
+		    if (read(0, &ch, 1) != 1)
+		    {
+			/* read error: either 0 or -1 */
+			new_message(MT_standout, " Read error on stdin");
+			putchar('\r');
+			quit(1);
+			/*NOTREACHED*/
+		    }
 		    if ((iptr = strchr(command_chars, ch)) == NULL)
 		    {
 			/* illegal command */
@@ -675,199 +881,221 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 			{
 			    /* error */
 			    new_message(MT_standout,
-			    " Command cannot be handled by this terminal");
+					" Command cannot be handled by this terminal");
 			    putchar('\r');
 			    no_command = Yes;
 			}
 			else switch(change)
 			{
-			    case CMD_redraw:	/* redraw screen */
-				reset_display();
-				break;
+			case CMD_redraw:	/* redraw screen */
+			    reset_display();
+			    break;
     
-			    case CMD_update:	/* merely update display */
-				/* is the load average high? */
-				if (system_info.load_avg[0] > LoadMax)
-				{
-				    /* yes, go home for visual feedback */
-				    go_home();
-				    fflush(stdout);
-				}
-				break;
+			case CMD_update:	/* merely update display */
+			    /* go home for visual feedback */
+			    go_home();
+			    fflush(stdout);
+			    break;
 	    
-			    case CMD_quit:	/* quit */
-				quit(0);
-				/*NOTREACHED*/
-				break;
+			case CMD_quit:	/* quit */
+			    quit(0);
+			    /*NOTREACHED*/
+			    break;
 	    
-			    case CMD_help1:	/* help */
-			    case CMD_help2:
+			case CMD_help1:	/* help */
+			case CMD_help2:
+			    reset_display();
+			    display_pagerstart();
+			    show_help(&statics);
+			    display_pagerend();
+			    break;
+	
+			case CMD_errors:	/* show errors */
+			    if (error_count() == 0)
+			    {
+				new_message(MT_standout,
+					    " Currently no errors to report.");
+				putchar('\r');
+				no_command = Yes;
+			    }
+			    else
+			    {
 				reset_display();
 				clear();
-				show_help();
+				show_errors();
 				standout("Hit any key to continue: ");
 				fflush(stdout);
 				(void) read(0, &ch, 1);
-				break;
+			    }
+			    break;
 	
-			    case CMD_errors:	/* show errors */
-				if (error_count() == 0)
+			case CMD_number1:	/* new number */
+			case CMD_number2:
+			    new_message(MT_standout,
+					"Number of processes to show: ");
+			    newval = readline(tempbuf1, 8, Yes);
+			    if (newval > -1)
+			    {
+				if (newval > max_topn)
+				{
+				    new_message(MT_standout | MT_delayed,
+						" This terminal can only display %d processes.",
+						max_topn);
+				    putchar('\r');
+				}
+
+				if (newval == 0)
+				{
+				    /* inhibit the header */
+				    display_header(No);
+				}
+				else if (newval > topn && topn == 0)
+				{
+				    /* redraw the header */
+				    display_header(Yes);
+				    d_header = i_header;
+				}
+				topn = newval;
+			    }
+			    break;
+	    
+			case CMD_delay:	/* new seconds delay */
+			    new_message(MT_standout, "Seconds to delay: ");
+			    if ((i = readline(tempbuf1, 8, Yes)) > -1)
+			    {
+				if ((delay = i) == 0 && getuid() != 0)
+				{
+				    delay = 1;
+				}
+			    }
+			    clear_message();
+			    break;
+	
+			case CMD_displays:	/* change display count */
+			    new_message(MT_standout,
+					"Displays to show (currently %s): ",
+					displays == -1 ? "infinite" :
+					itoa(displays));
+			    if ((i = readline(tempbuf1, 10, Yes)) > 0)
+			    {
+				displays = i;
+			    }
+			    else if (i == 0)
+			    {
+				quit(0);
+			    }
+			    clear_message();
+			    break;
+    
+#ifdef ENABLE_KILL
+			case CMD_kill:	/* kill program */
+			    new_message(0, "kill ");
+			    if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
+			    {
+				if ((errmsg = kill_procs(tempbuf2)) != NULL)
+				{
+				    new_message(MT_standout, "%s", errmsg);
+				    putchar('\r');
+				    no_command = Yes;
+				}
+			    }
+			    else
+			    {
+				clear_message();
+			    }
+			    break;
+	    
+			case CMD_renice:	/* renice program */
+			    new_message(0, "renice ");
+			    if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
+			    {
+				if ((errmsg = renice_procs(tempbuf2)) != NULL)
+				{
+				    new_message(MT_standout, "%s", errmsg);
+				    putchar('\r');
+				    no_command = Yes;
+				}
+			    }
+			    else
+			    {
+				clear_message();
+			    }
+			    break;
+#endif
+
+			case CMD_idletog:
+			case CMD_idletog2:
+			    ps.idle = !ps.idle;
+			    new_message(MT_standout | MT_delayed,
+					" %sisplaying idle processes.",
+					ps.idle ? "D" : "Not d");
+			    putchar('\r');
+			    break;
+
+			case CMD_cmdline:
+			    if (statics.flags.fullcmds)
+			    {
+				ps.fullcmd = !ps.fullcmd;
+				new_message(MT_standout | MT_delayed,
+					    " %sisplaying full command lines.",
+					    ps.fullcmd ? "D" : "Not d");
+			    }
+			    else
+			    {
+				new_message(MT_standout, " Full command display not supported.");
+				no_command = Yes;
+			    }
+			    putchar('\r');
+			    break;
+	    
+			case CMD_user:
+			    new_message(MT_standout,
+					"Username to show: ");
+			    if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
+			    {
+				if (tempbuf2[0] == '+' &&
+				    tempbuf2[1] == '\0')
+				{
+				    ps.uid = -1;
+				}
+				else if ((i = userid(tempbuf2)) == -1)
 				{
 				    new_message(MT_standout,
-					" Currently no errors to report.");
-				    putchar('\r');
+						" %s: unknown user", tempbuf2);
 				    no_command = Yes;
 				}
 				else
 				{
-				    reset_display();
-				    clear();
-				    show_errors();
-				    standout("Hit any key to continue: ");
-				    fflush(stdout);
-				    (void) read(0, &ch, 1);
+				    ps.uid = i;
 				}
-				break;
-	
-			    case CMD_number1:	/* new number */
-			    case CMD_number2:
-				new_message(MT_standout,
-				    "Number of processes to show: ");
-				newval = readline(tempbuf1, 8, Yes);
-				if (newval > -1)
-				{
-				    if (newval > max_topn)
-				    {
-					new_message(MT_standout | MT_delayed,
-					  " This terminal can only display %d processes.",
-					  max_topn);
-					putchar('\r');
-				    }
-
-				    if (newval == 0)
-				    {
-					/* inhibit the header */
-					display_header(No);
-				    }
-				    else if (newval > topn && topn == 0)
-				    {
-					/* redraw the header */
-					display_header(Yes);
-					d_header = i_header;
-				    }
-				    topn = newval;
-				}
-				break;
-	    
-			    case CMD_delay:	/* new seconds delay */
-				new_message(MT_standout, "Seconds to delay: ");
-				if ((i = readline(tempbuf1, 8, Yes)) > -1)
-				{
-				    delay = i;
-				}
-				clear_message();
-				break;
-	
-			    case CMD_displays:	/* change display count */
-				new_message(MT_standout,
-					"Displays to show (currently %s): ",
-					displays == -1 ? "infinite" :
-							 itoa(displays));
-				if ((i = readline(tempbuf1, 10, Yes)) > 0)
-				{
-				    displays = i;
-				}
-				else if (i == 0)
-				{
-				    quit(0);
-				}
-				clear_message();
-				break;
-    
-			    case CMD_kill:	/* kill program */
-				new_message(0, "kill ");
-				if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
-				{
-				    if ((errmsg = kill_procs(tempbuf2)) != NULL)
-				    {
-					new_message(MT_standout, errmsg);
-					putchar('\r');
-					no_command = Yes;
-				    }
-				}
-				else
-				{
-				    clear_message();
-				}
-				break;
-	    
-			    case CMD_renice:	/* renice program */
-				new_message(0, "renice ");
-				if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
-				{
-				    if ((errmsg = renice_procs(tempbuf2)) != NULL)
-				    {
-					new_message(MT_standout, errmsg);
-					putchar('\r');
-					no_command = Yes;
-				    }
-				}
-				else
-				{
-				    clear_message();
-				}
-				break;
-
-			    case CMD_idletog:
-			    case CMD_idletog2:
-				ps.idle = !ps.idle;
-				new_message(MT_standout | MT_delayed,
-				    " %sisplaying idle processes.",
-				    ps.idle ? "D" : "Not d");
 				putchar('\r');
-				break;
+			    }
+			    else
+			    {
+				clear_message();
+			    }
+			    break;
 
-			    case CMD_user:
+			case CMD_order:
+			    if (statics.order_names == NULL)
+			    {
+				new_message(MT_standout, " Ordering not supported.");
+				putchar('\r');
+				no_command = Yes;
+			    }
+			    else
+			    {
 				new_message(MT_standout,
-				    "Username to show: ");
+					    "Order to sort: ");
 				if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
 				{
-				    if (tempbuf2[0] == '+' &&
-					tempbuf2[1] == '\0')
-				    {
-					ps.uid = -1;
-				    }
-				    else if ((i = userid(tempbuf2)) == -1)
+				    if ((i = string_index(tempbuf2, statics.order_names)) == -1)
 				    {
 					new_message(MT_standout,
-					    " %s: unknown user", tempbuf2);
+						    " %s: unrecognized sorting order", tempbuf2);
 					no_command = Yes;
 				    }
 				    else
-				    {
-					ps.uid = i;
-				    }
-				    putchar('\r');
-				}
-				else
-				{
-				    clear_message();
-				}
-				break;
-	    
-#ifdef ORDER
-			    case CMD_order:
-				new_message(MT_standout,
-				    "Order to sort: ");
-				if (readline(tempbuf2, sizeof(tempbuf2), No) > 0)
-				{
-				  if ((i = string_index(tempbuf2, statics.order_names)) == -1)
-					{
-					  new_message(MT_standout,
-					      " %s: unrecognized sorting order", tempbuf2);
-					  no_command = Yes;
-				    }
-				    else
 				    {
 					order_index = i;
 				    }
@@ -877,12 +1105,81 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
 				{
 				    clear_message();
 				}
-				break;
-#endif
-	    
-			    default:
-				new_message(MT_standout, " BAD CASE IN SWITCH!");
+			    }
+			    break;
+                        case CMD_order_pid:
+                            if ((i = string_index("pid", statics.order_names)) == -1) {
+                                new_message(MT_standout,
+                                            " Unrecognized sorting order");
 				putchar('\r');
+			        no_command = Yes;
+                            } else {
+                                order_index = i;
+                            }
+			    break;
+                        case CMD_order_cpu:
+                            if ((i = string_index("cpu", statics.order_names)) == -1) {
+                                new_message(MT_standout,
+                                            " Unrecognized sorting order");
+				putchar('\r');
+			        no_command = Yes;
+                            } else {
+                                order_index = i;
+                            }
+			    break;
+                        case CMD_order_mem:
+                            if ((i = string_index("size", statics.order_names)) == -1) {
+                                new_message(MT_standout,
+                                            " Unrecognized sorting order");
+				putchar('\r');
+			        no_command = Yes;
+                            } else {
+                                order_index = i;
+                            }
+			    break;
+                        case CMD_order_time:
+                            if ((i = string_index("time", statics.order_names)) == -1) {
+                                new_message(MT_standout,
+                                            " Unrecognized sorting order");
+				putchar('\r');
+			        no_command = Yes;
+                            } else {
+                                order_index = i;
+                            }
+			    break;
+
+#ifdef ENABLE_COLOR
+			case CMD_color:
+			    reset_display();
+			    if (color_on)
+			    {
+				color_on = 0;
+				display_resize();  /* To realloc screenbuf */
+				new_message(MT_standout | MT_delayed,
+					    " Color off");
+			    }
+			    else
+			    {
+				if (!smart_terminal)
+				{
+				    new_message(MT_standout | MT_delayed,
+						" Sorry, cannot do colors on this terminal type");
+				}
+				else
+				{
+				    color_on = 1;
+				    new_message(MT_standout | MT_delayed,
+						" Color on");
+				}
+			    }
+			    
+			    break;
+#endif
+
+			default:
+			    new_message(MT_standout, " Unsupported command");
+			    putchar('\r');
+			    no_command = Yes;
 			}
 		    }
 
@@ -897,100 +1194,3 @@ Usage: %s [-ISbinqu] [-d x] [-s x] [-o field] [-U username] [number]\n",
     /*NOTREACHED*/
 }
 
-/*
- *  reset_display() - reset all the display routine pointers so that entire
- *	screen will get redrawn.
- */
-
-reset_display()
-
-{
-    d_loadave    = i_loadave;
-    d_procstates = i_procstates;
-    d_cpustates  = i_cpustates;
-    d_memory     = i_memory;
-    d_message	 = i_message;
-    d_header	 = i_header;
-    d_process	 = i_process;
-}
-
-/*
- *  signal handlers
- */
-
-sigret_t leave()	/* exit under normal conditions -- INT handler */
-
-{
-    end_screen();
-    exit(0);
-}
-
-sigret_t tstop(i)	/* SIGTSTP handler */
-
-int i;
-
-{
-    /* move to the lower left */
-    end_screen();
-    fflush(stdout);
-
-    /* default the signal handler action */
-    (void) signal(SIGTSTP, SIG_DFL);
-
-    /* unblock the signal and send ourselves one */
-#ifdef SIGRELSE
-    sigrelse(SIGTSTP);
-#else
-    (void) sigsetmask(sigblock(0) & ~(1 << (SIGTSTP - 1)));
-#endif
-    (void) kill(0, SIGTSTP);
-
-    /* reset the signal handler */
-    (void) signal(SIGTSTP, tstop);
-
-    /* reinit screen */
-    reinit_screen();
-
-    /* jump to appropriate place */
-    longjmp(jmp_int, 1);
-
-    /*NOTREACHED*/
-}
-
-#ifdef SIGWINCH
-sigret_t winch(i)		/* SIGWINCH handler */
-
-int i;
-
-{
-    /* reascertain the screen dimensions */
-    get_screensize();
-
-    /* tell display to resize */
-    max_topn = display_resize();
-
-    /* reset the signal handler */
-    (void) signal(SIGWINCH, winch);
-
-    /* jump to appropriate place */
-    longjmp(jmp_int, 1);
-}
-#endif
-
-void quit(status)		/* exit under duress */
-
-int status;
-
-{
-    end_screen();
-    exit(status);
-    /*NOTREACHED*/
-}
-
-sigret_t onalrm()	/* SIGALRM handler */
-
-{
-    /* this is only used in batch mode to break out of the pause() */
-    /* return; */
-}
-
